package raytracer;

import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.DisplayMode;
import org.lwjgl.opengl.GL11;

import raytracer.graphics.Shader;

public class RayTracer {
	
	public final int width, height;
	private Camera camera;
	
	private RayTracer(int width, int height) {
		this.width = width;
		this.height = height;
		
		initGL();
		Shader.init();
		camera = new Camera(70f, 16f / 9f, 0.02f, 1000f);
		camera.loadInvProjMatrToShader(Shader.rayShader);
		Shader.rayShader.setUniform2f("windowSize", width, height);
	}
	
	private void tick() {
		camera.update();
	}
	
	private void render() {
		Shader.rayShader.enable();
		GL11.glRectf(-1, -1, 1, 1);
	}
	
	private void loop() {
		long prevTime = System.nanoTime(), currentTime, realPrevTime = prevTime, realDeltaTime = 0;
		int delta, ticks = 0, frames = 0, fps = 0;
		while (!Display.isCloseRequested()) {
			delta = (int) (((currentTime = System.nanoTime()) - prevTime) * 1e-9 * 60);
			realPrevTime = currentTime;
			for (int i = 0; i < delta; i++) {
				tick();
				ticks++;
				if (ticks % 60 == 0) {
					fps = frames;
					frames = 0;
				}
				prevTime += 1e9 / 60;
			}
			
			GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);
			render();
			
			int error = GL11.glGetError();
			if (error != GL11.GL_NO_ERROR) System.out.println("OpenGL error: " + error);
			
			String debugInfo = "    P: (" + camera.position.x + ", " + camera.position.y + ", " + camera.position.z + ")    R: (" + (float) (camera.rotation.x * 180 / Math.PI) + ", " + (float) (camera.rotation.y * 180 / Math.PI) + ", " + (float) (camera.rotation.z * 180 / Math.PI) + ")";
			Display.update();
			realDeltaTime = System.nanoTime() - realPrevTime;
			frames++;
			Display.setTitle(fps + " fps   ms: " + ((int) (realDeltaTime * 1e-6) < 10 ? "0" : "") + (int) (realDeltaTime * 1e-6) + debugInfo);
			Display.sync(60);
		}
	}
	
	private void initGL() {
		try {
			Display.setDisplayMode(new DisplayMode(width, height));
			Display.create();
//			Keyboard.create();
//			Mouse.create();
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		GL11.glClearColor(0, 0, 0, 1);
		GL11.glEnable(GL11.GL_DEPTH_TEST);
	}
	
	public static void destroy() {
		Display.destroy();
	}
	
	public static void main(String[] args) {
		String os = System.getProperty("os.name").toLowerCase();
		String path;
		if (os.contains("win")) path = "windows";
		else if (os.contains("mac")) path = "macosx";
		else if (os.contains("sol")) path = "solaris";
		else path = "linux";
		String libPath = RayTracer.class.getResource("/native/" + path).getPath();
		if (libPath != null) System.setProperty("org.lwjgl.librarypath", libPath);
		
		RayTracer rt = new RayTracer(1280, 720);
		rt.loop();
		RayTracer.destroy();
	}
	
}
